
#ifndef SANGOMA_INTERFACE_H
#define SANGOMA_INTERFACE_H

#include <stdio.h>
#include <stddef.h>		//for offsetof()
#include <stdlib.h>

#if defined(__WINDOWS__)
# include <windows.h>
# include <winioctl.h>
# include <conio.h>
# include "bit_win.h"
# include "wanpipe_time.h" //for wp_usleep()

#elif defined(__LINUX__)

/* Include headers */
# include <stddef.h>
# include <errno.h>
# include <fcntl.h>
# include <string.h>
# include <ctype.h>
# include <sys/stat.h>
# include <sys/ioctl.h>
# include <sys/types.h>
# include <dirent.h>
# include <unistd.h>
# include <sys/socket.h>
# include <netdb.h>
# include <sys/un.h>
# include <sys/wait.h>
# include <unistd.h>
# include <signal.h>
# include <time.h>
#else
# error "sangoma_interface.h: undefined OS type"
#endif

#include "wanpipe_api.h"
#include "libsangoma.h"

#define SNG_IF_NAME_LEN 100

/*!
  \class sangoma_interface
  \brief Sangoma Interface Class that deals with span/chan IO (read/write/events)
*/
class sangoma_interface
{
protected:
	/*! Sangoma IO device descriptor */
	sng_fd_t	sangoma_dev;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	/*! deprecated: ioctl management structure, used to execute management commands - one should use libsangoma directly */
	wan_udp_hdr_t	wan_udp;

	/*! deprecated: Helper function to get data byte out of wan_udp structure */
	unsigned char get_wan_udphdr_data_byte(unsigned char off);

	/*! deprecated: Helper function to get data pointer out of wan_udp structure */
	unsigned char *get_wan_udphdr_data_ptr(unsigned char off);

	/*! deprecated: Helper function to set data pointer out of wan_udp structure */
	unsigned char set_wan_udphdr_data_byte(unsigned char off, unsigned char data);

	/*! device low level protocol (T1/E1...) currently running  */
	int 			wan_protocol;
    /*! device adapter type  */
	unsigned char	adapter_type;

	/*! deprecated: API command structure used to execute API commands  */
	wanpipe_api_cmd_t tdm_api_cmd;

	/*! API command structure used to execute API commands. This command structure is used with libsangoma library */
	wanpipe_api_t wp_api;
	
	/*! Shutdown function to cleanup the class */
	void cleanup();

	/*! Get device span configuration */
	int get_wan_config();

	/*! Get interface chan configuration */
	int get_interface_configuration(if_cfg_t *wanif_conf_ptr);

	/*! Get front end type (T1/E1...) */
	int get_fe_type(unsigned char* adapter_type);

	/*! deprecated: Execute API command IOCTL - one should use libsangoma directly */
	int wanpipe_api_ioctl(wan_cmd_api_t *api_cmd);

	/*! Wanpipe Interface configuration structure to be populated on read configuration */
	if_cfg_t				wanif_conf_struct;

	/*! Wanpipe Device number based on hardware probe device enumeration (SPAN) */
	int WanpipeNumber;

	/*! Wanpipe Interface number based on a channel in a SPAN */
	int InterfaceNumber;

public:
	char	device_name[SNG_IF_NAME_LEN];

	//////////////////////////////////////////////////////////////////
	//methods
	sangoma_interface(int wanpipe_number, int interface_number);
	~sangoma_interface();

	int DoManagementCommand(sng_fd_t drv, wan_udp_hdr_t* packet);

	virtual int init();
	
	int print_api_info();
	int print_channels_info();

	int get_open_handles_counter();

	sng_fd_t open_api_device();

	unsigned char get_adapter_type();
	unsigned int get_sub_media();

	int tdmv_api_ioctl(wanpipe_api_cmd_t *api_cmd);

	int reset_interface_state();
};

#endif//SANGOMA_INTERFACE_H

