//////////////////////////////////////////////////////////////////////
// sangoma_port.cpp: implementation of the sangoma_port class.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "sangoma_port.h"

#if 0
# define DBG_SNG_PORT	printf("%s():line:%d: ", __FUNCTION__, __LINE__);printf
# define _DBG_SNG_PORT	printf
#else
# define DBG_SNG_PORT
# define _DBG_SNG_PORT
#endif

#define ERR_SNG_PORT	printf("Error: %s():line:%d: ", __FUNCTION__, __LINE__);printf
#define _ERR_SNG_PORT	printf

#define SNG_PORT_FUNC()	DBG_SNG_PORT("\n")	

#pragma comment( lib, "libsangoma" )	/* import functions from libsangoma.lib */

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

sangoma_port::sangoma_port()
{
	SNG_PORT_FUNC();
	wp_handle = INVALID_HANDLE_VALUE;
}

sangoma_port::~sangoma_port()
{
	SNG_PORT_FUNC();
	cleanup();
}

//////////////////////////////////////////////////////////////////////
// other functions
//////////////////////////////////////////////////////////////////////

int sangoma_port::init(uint16_t wanpipe_number)
{
//	SNG_PORT_FUNC();
	wp_number = wanpipe_number;
	wp_handle = sangoma_open_driver_ctrl(wanpipe_number);

	if(wp_handle == INVALID_HANDLE_VALUE){
		//ERR_SNG_PORT("failed to open wanpipe%d!!\n", wp_number);
		return 1;
	}

	return 0;
}

void sangoma_port::cleanup()
{
	if(wp_handle != INVALID_HANDLE_VALUE){
		sangoma_close(&wp_handle);
		wp_handle = INVALID_HANDLE_VALUE;
	}
}

int sangoma_port::get_hardware_info(hardware_info_t *hardware_info)
{
	port_management_struct_t port_mgmnt;
	int rc;
	
	memset(&port_mgmnt, 0x00, sizeof(port_mgmnt));

	rc = sangoma_driver_get_hw_info(wp_handle, &port_mgmnt, wp_number);
	if (rc) {
		ERR_SNG_PORT("sangoma_driver_get_hw_info() failed for wanpipe%d!\n", wp_number);
		return rc;
	}

	memcpy(hardware_info, port_mgmnt.data, sizeof(*hardware_info));

	return 0;
}

int sangoma_port::start_port()
{
	port_management_struct_t port_mgmnt;
	int rc;

	memset(&port_mgmnt, 0x00, sizeof(port_mgmnt));

	rc = sangoma_driver_port_start(wp_handle,&port_mgmnt,wp_number);
	if(rc){
		WANCFG_BREAK_POINT();
		DBG_SNG_PORT("rc: %s (%d)\n", SDLA_DECODE_SANG_STATUS(rc), rc);
		return rc;
	}

	return port_mgmnt.operation_status;
}

int sangoma_port::stop_port()
{
	port_management_struct_t port_mgmnt;
	int rc;

	memset(&port_mgmnt, 0x00, sizeof(port_mgmnt));

	rc = sangoma_driver_port_stop(wp_handle, &port_mgmnt, wp_number);
	if(rc){
		WANCFG_BREAK_POINT();
		return rc;
	}

	if(port_mgmnt.operation_status){
		if(SANG_STATUS_CAN_NOT_STOP_DEVICE_WHEN_ALREADY_STOPPED == port_mgmnt.operation_status){
			port_mgmnt.operation_status = SANG_STATUS_SUCCESS;
		}else{
			WANCFG_BREAK_POINT();
		}
	}

	return port_mgmnt.operation_status;
}

sng_fd_t sangoma_port::get_wp_handle()
{
	return this->wp_handle;
}
