// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently, but
// are changed infrequently
//

#pragma once

#include "targetver.h"

#define _CRTDBG_MAP_ALLOC //enable memory leaks checking

#ifndef WIN32_LEAN_AND_MEAN
# define WIN32_LEAN_AND_MEAN	// Exclude rarely-used stuff from Windows headers
#endif

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <crtdbg.h> 

//////////////////////////////////////////////////////////////////
// Definitions and Declarations private for this DLL.
//////////////////////////////////////////////////////////////////
#define WP_MAX_PORTS		(16*8)	/* 16 A108 cards is a reasonable maximum */
#define WP_MAX_INTERFACES	(32)	/* maximum number of interfaces on a port */

#define WP_API_CFG_FORMAT	"%-16s%-15s%-12s%-12s%-20s\n"
#define WP_CHAN_CFG_FORMAT	"%-16s%-12s%-12s%-12s\n"

#include "wanpipe_includes.h"
#include "wanpipe_cfg.h"	//for port_cfg_t
/***************************************************************************************/
/* Private functions.
 * Note: some of these functions can be exported */
#ifdef __cplusplus
extern "C" {	/* for C++ users */
#endif

int wancfglib_get_used_by_integer_value(char *used_by_str);
int wancfglib_get_config_id_from_profile(char *config_id);
unsigned int wancfglib_parse_active_channel(char *val);
char* wancfglib_get_protocol_string(int protocol);
char* wancfglib_get_used_by_string(int used_by);
char* wancfglib_remove_spaces_in_int_string(char *input);
int wancfglib_config_id_to_annexg_id(int config_id);
int wancfglib_print_port_cfg_structure(port_cfg_t *port_cfg);

#ifdef __cplusplus
}	/* for C++ users */
#endif
/***************************************************************************************/


#define LEN_OF_DBG_BUFF   4096

/* Flag used around variabales used ONLY for debugging.
 * Set to zero in "release" version to avoid 'unused variable' message
 * from the compiler. */
#if 0
# define DBG_FLAG 1
#else
# define DBG_FLAG 0
#endif

#if DBG_FLAG
# define Debug(dbg_flag, x)		if(dbg_flag)printf x;
# define WANCFG_FUNC_DBG()		printf("%s(), Line: %d\n", __FUNCTION__, __LINE__)

# define WANCFG_BREAK_POINT()	DebugBreak()
//# define WANCFG_BREAK_POINT()	printf("BREAK_POINT in %s() at Line: %d\n", __FUNCTION__, __LINE__)

#else
# define Debug(dbg_flag, x)
# define WANCFG_FUNC_DBG()
#define WANCFG_BREAK_POINT()
#endif

static void DebugV(void *pszFormat, ...)
{
#if DBG_FLAG
	unsigned char str[400];
	va_list	vaArg;

	va_start (vaArg, pszFormat);
	_vsnprintf((char *)str, sizeof(str), (char *)pszFormat, vaArg);
	va_end (vaArg);

#if 0
	//to stdout
	printf( (char*)str );
#else
	//to debugger
	OutputDebugStringA( (char*)str );
#endif

#endif//DBG_FLAG
}

#define INVALID_CARD_TYPE	0
#define NOT_SET				100 /* can be stored in a variable as small as 1 byte */
#define PROTOCOL_NOT_SET	0
#define NO_PROTOCOL_NEEDED  (WANCONFIG_X25 - 1)
#define INVALID_PROTOCOL	"Invalid Protocol!!"

/* The TDM_VOICE and TDM_VOICE_API is actually an Operation Mode (like API...)
 * but there is a need to display and configure it as a Protocol.
 * No such 'config_id' in wanpipe_cfg.h', so have to define it here. */
enum NOT_REAL_PROTOCOLS{
	PROTOCOL_TDM_VOICE = 60,
	PROTOCOL_IP,
	PROTOCOL_ETHERNET,
	PROTOCOL_TDM_VOICE_API,
	PROTOCOL_TDM_SPAN_VOICE_API,
	PROTOCOL_TDM_CHAN_VOICE_API,
	PROTOCOL_API_LEGACY
};

#define PROGRAM_NAME "wancfglib.dll"

/* macro to display a critical error and to exit the program */
#define ERR_DBG_OUT(_x_)	\
{	printf("%s: Error in Function: %s(), at Line: %d.\n Text: ",	\
		PROGRAM_NAME, __FUNCTION__, __LINE__);	\
	printf _x_;	\
	WANCFG_BREAK_POINT();	\
}

#define WANCFGLIB_MALLOC(type, buffer, length)	\
{	\
	buffer = (##type*)malloc(length);	\
	if(!buffer){	\
		printf("Error: %s(): failed to allocated %d bytes of memory!\n", __FUNCTION__, length);	\
	}else{	\
		memset(buffer, 0x00, length);	\
	}	\
}
