/***************************************************************************
wancfglib_conf_file_reader.cpp  -  description
-------------------
begin                : Thu Mar 4 2004
copyright            : (C) 2004 by David Rokhvarg
email                : davidr@sangoma.com
***************************************************************************/

/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************/

#include "stdafx.h"
#include "wancfglib_conf_file_reader.h"

#define DBG_CONF_FILE_READER	  0
#define DBG_CONF_FILE_READER_AFT  0

#define USE_ANNEXG_TABLE	1

static void SIZEOFASSERT (key_word_t* dtab, unsigned int type_size)
{
	if (dtab->size != type_size) {
		fprintf(stderr,"\n==========CRITICAL ERROR============\n\n");
		fprintf(stderr,"Size Mismatch: Type Size %i != %i\n",dtab->size, type_size);
		fprintf(stderr,"======================================\n\n");
		fprintf(stderr,"Plese email /var/log/wanrouter file to Sangoma Support\n");
		fprintf(stderr,"Email to techdesk@sangoma.com\n");
		fprintf(stderr,"======================================\n\n");
	}
}

/****** Global Function Prototypes *************************************************/

int tokenize (char* str, char **tokens);
char* str_strip (char* str, char* s);
char* strupcase	(char* str);
void* lookup (int val, look_up_t* table);
int name2val (char* name, look_up_t* table);
unsigned int dec_to_uint (unsigned char* str, int len);
void read_adsl_vci_vpi_list(wan_adsl_vcivpi_t* vcivpi_list, unsigned short* vcivpi_num);

//////////////////////////////////////////////////////////////////////////////////////////
//class member functions
//////////////////////////////////////////////////////////////////////////////////////////

wancfglib_conf_file_reader::wancfglib_conf_file_reader(int wp_number)
{
	Debug(DBG_CONF_FILE_READER, ("%s(%d)\n", __FUNCTION__, wp_number));

	wanpipe_number = wp_number;

	snprintf(link_defs.name, sizeof(link_defs.name), "wanpipe%d", wanpipe_number);
#if DBG_FLAG
	verbose = 1;   /* verbosity level */
#else
	verbose = 0;   /* verbosity level */
#endif
	conf_file_ptr = NULL;

	memset(wanpipe_cfg_dir, 0x00, sizeof(wanpipe_cfg_dir));
}

wancfglib_conf_file_reader::~wancfglib_conf_file_reader()
{
	Debug(DBG_CONF_FILE_READER, ("%s()\n", __FUNCTION__));

	if(conf_file_ptr != NULL){
		fclose(conf_file_ptr);
	}
}

int wancfglib_conf_file_reader::get_wanpipe_cfg_directory()
{
	char cSystemWindowsDir[MAX_PATH];
	char separators[] = ":";//a separator in something like "C:\WINDOWS"
	char *token;
	int token_counter = 0;

	if(!GetSystemWindowsDirectory(cSystemWindowsDir, MAX_PATH)){
		ERR_DBG_OUT(("GetSystemWindowsDirectory() failed!\n"));		
		return 1;
	}

	//the first token:
	token = strtok( cSystemWindowsDir, separators );
	//We are interested ONLY in the 1-st token from something 
	//like "C:\WINDOWS", which is "C".
	if( token != NULL ){
		snprintf(wanpipe_cfg_dir, sizeof(wanpipe_cfg_dir), "%s:\\Sangoma\\", token);
	}else{
		//error
		return 2;
	}

	return 0;
}

int wancfglib_conf_file_reader::read_conf_file()
{
	int err = 0;

	if(get_wanpipe_cfg_directory()){
		ERR_DBG_OUT(("Failed to get Sangoma Configuration Directory!\n"));		
	}

	snprintf(conf_file_full_path, sizeof(conf_file_full_path), "%swanpipe%d.conf", wanpipe_cfg_dir, wanpipe_number);

	conf_file_ptr = fopen(conf_file_full_path, "r");
	if (conf_file_ptr == NULL) {
		ERR_DBG_OUT(("Failed to open file: %s!!\n", conf_file_full_path));
	}

	err = read_devices_section(conf_file_ptr);
	if (err){
		return err;
	}

	err = read_wanpipe_section(conf_file_ptr);
	if (err){
		return err;
	}

#if DBG_CONF_FILE_READER
	sdla_fe_cfg_t	*fe_cfg = &link_defs.linkconf->fe_cfg;
	Debug(DBG_CONF_FILE_READER, ("%s(): fe_cfg->media: 0x%x\n",__FUNCTION__, fe_cfg->media));
	Debug(DBG_CONF_FILE_READER, ("%s(): fe_cfg->line_no: %d\n",__FUNCTION__, fe_cfg->line_no));
#endif

	int current_line_number = 0;
	err = read_interfaces_section(
					conf_file_ptr,		//FILE* file,
					*this,				//objects_list* parent_objects_list,
					NULL,				//char* parent_dev_name,
					&current_line_number//int* current_line_number
					);
	
	if(!err){
		link_defs.config_id = link_defs.linkconf->config_id;
	}

	return err;
}


int wancfglib_conf_file_reader::get_port_cfg(port_cfg_t	*port_cfg)
{
	memset(port_cfg, 0x00, sizeof(*port_cfg));

	//first get the wandev_conf_t part
	wandev_conf_t *wandev_conf = link_defs.linkconf;

	Debug(DBG_CONF_FILE_READER, ("%s(): wandev_conf->config_id: %d\n", __FUNCTION__, wandev_conf->config_id));

	memcpy(&port_cfg->wandev_conf, wandev_conf, sizeof(*wandev_conf));

	//now get configuration of each interface
	port_cfg->num_of_ifs = next_objects_list.size();

	for(unsigned int index = 0; index < next_objects_list.size(); ++index){

		wanif_conf_t *wanif_conf = next_objects_list.at(index)->if_data.chanconf;

		//For the AFT interfaces with config id such as WAN_AFT_TE1 the 'wanif_conf->mtu'
		//is not initialized, but instead, 'wanif_conf->u.aft.mtu' is initialized.
		//That means we must do it here!
		wanif_conf->mtu = wanif_conf->u.aft.mru = wanif_conf->u.aft.mtu;

		memcpy(&port_cfg->if_cfg[index], wanif_conf, sizeof(*wanif_conf));
	}

#if DBG_CONF_FILE_READER
	wancfglib_print_port_cfg_structure(port_cfg);
#endif

	return 0;
}

/* Currently the only special action of hw interfaces is
 * HWEC configuration. */
int wancfglib_conf_file_reader::configure_hw_interfaces()
{
	int rc;

	for(unsigned int index = 0; index < next_objects_list.size(); ++index){

		rc = wanconfig_hwec(&next_objects_list.at(index)->if_data);
		if(rc){
			WANCFG_BREAK_POINT();
			return rc;
		}
	}

	return 0;
}

/*============================================================================
* Functions for parsing a wanpipe?.conf file.
*/

int wancfglib_conf_file_reader::read_devices_section(FILE* file)
{
	int err = 0;
	char* conf;		  /* -> section buffer */
	char* key;		  /* -> current configuration record */
	int len;		  /* record length */
	char *token[MAX_TOKENS];

	int toknum, token_index;

	Debug(DBG_CONF_FILE_READER, ("%s()\n", __FUNCTION__));

	/* Read [devices] section */
	conf = read_conf_section(file, "devices");
	if (conf == NULL){
		return ERR_CONFIG;
	}

	/* For the line in [devices] section create a link definition structure. */
	/* Only a SINGLE line allowed in [devices] section. */
	/* The line will be similar to: "wanpipe1 = WAN_AFT_TE1, Comment" */
	key = conf;
	len = strlen(key) + 1;

	toknum = tokenize(key, token);
	if(toknum < 2){
		ERR_DBG_OUT(("Invalid number of tokens (%d) in [devices] section!!\n", toknum));
		err = ERR_CONFIG;
		goto done;
	}

	for(token_index = 0; token_index < toknum; token_index++){

		switch(token_index)
		{
		case 0:
			//store the device name. i.g. wanpipe1
			strncpy(link_defs.name, token[0], WAN_DRVNAME_SZ);
			break;
		case 1:
			//store config id. i.g. WAN_AFT_TE1.
			strupcase(token[1]);
			link_defs.linkconf->config_id = name2val(token[1], config_id_str);
			if (!link_defs.linkconf->config_id) {
				ERR_DBG_OUT((" * Device config_id %s is invalid!\n", token[1]));
				err = ERR_CONFIG;
				goto done;
			}
			Debug(DBG_CONF_FILE_READER, ("link_defs.linkconf->config_id: %d\n",
				link_defs.linkconf->config_id));
			break;
		case 2:
			if ((toknum > 2) && token[2]){
				link_defs.descr = strdup(token[2]);
			}
			break;
		default:
			printf("Warning: unused token in [devices] section: %s\n", token[token_index]);
			break;
		}
	}//for()

done:
	Debug(DBG_CONF_FILE_READER, ("%s() - End (err:%d)\n", __FUNCTION__, err));
	free(conf);
	return err;
}

/*============================================================================
* 
*/
int wancfglib_conf_file_reader::read_wanpipe_section(FILE* file)
{
	int err = 0;
	int len = 0;
	key_word_t* conf_table = (key_word_t*)lookup(link_defs.linkconf->config_id, conf_def_tables);
	char* conf_rec;
	char *token[MAX_TOKENS];

	Debug(DBG_CONF_FILE_READER, ("%s()\n", __FUNCTION__));

	link_defs.conf = read_conf_section(file, link_defs.name);
	if(link_defs.conf == NULL){
		ERR_DBG_OUT(("Failed to read '%s' section!!\n", link_defs.name));
		err = ERR_CONFIG;
		return err;
	}

	/* Parse link configuration */
	for (conf_rec = link_defs.conf; *conf_rec; conf_rec += len) {

		int toknum;

		len = strlen(conf_rec) + 1;
		toknum = tokenize(conf_rec, token);

		Debug(DBG_CONF_FILE_READER, ("toknum: %d\n", toknum));

		if(toknum < 2) {
			ERR_DBG_OUT(("Invalid number of tokens (%d) in 'wanpipe' section!!\n", toknum));
			return ERR_CONFIG;
		}
		Debug(DBG_CONF_FILE_READER, ("token[0]: %s\n", token[0]));
		Debug(DBG_CONF_FILE_READER, ("token[1]: %s\n", token[1]));

		/* Look up a keyword first in common, then in media-specific
		* configuration definition tables. */
		strupcase(token[0]);
		//common		  key	    value
		err = set_conf_param(token[0], token[1], common_conftab, link_defs.linkconf, sizeof(wandev_conf_t));

		// 'config_id' specific
		if ((err < 0) && (conf_table != NULL)){
			err = set_conf_param(token[0], token[1], conf_table, &link_defs.linkconf->u, sizeof(link_defs.linkconf->u));
		}

		if(err < 0) {
			ERR_DBG_OUT((" * Unknown parameter %s!\n", token[0]));
			return ERR_CONFIG;
		}

		if (err){
			ERR_DBG_OUT(("Error parsing configuration file!\n"));
			return err;
		}
	}

	return err;
}

/*============================================================================
* Read configuration file section.
*	Return a pointer to memory filled with key entries read from given
*	section or NULL if section is not found. Each key is a zero-terminated
*	ASCII string with no leading spaces and comments. The last key is
*	followed by an empty string.
*	Note:	memory is allocated dynamically and must be released by the
*		caller.
*/
char* wancfglib_conf_file_reader::read_conf_section (FILE* file, char* section)
{
	char key[MAX_CFGLINE];		/* key buffer */
	char* buf = NULL;
	int found = 0, offs = 0, len, buf_len = 0;
	int dbg_len = 0;

	Debug(DBG_CONF_FILE_READER, ("%s(): reading section: %s\n", __FUNCTION__, section));
	Debug(DBG_CONF_FILE_READER, ("%s(): if_data ptr: 0x%p, if_data.addr: 0x%p\n", __FUNCTION__, &if_data, if_data.addr));

	rewind(file);

	while ((len = read_conf_record(file, key, MAX_CFGLINE)) > 0) {
		char* tmp;

		dbg_len += len;

		if (found) {

			if (*key == '['){
				break;	/* end of section */
			}

			if (buf){
				buf_len = offs+len+1;
				tmp = (char*)realloc(buf, offs + len + 1);
			}else{
				WANCFGLIB_MALLOC(char, tmp, len + 1);
				buf_len = len+1;
			}

			if (tmp) {
				buf = tmp;
				strlcpy(&buf[offs], key, buf_len-offs);
				offs += len;
				buf[offs] = '\0';
			}else{
				/* allocation failed */
				ERR_DBG_OUT(("Failed to allocate memory!\n"));
				break;
			}

		}else if (*key == '[') {
			//getting here on the 1-st pass	    
			tmp = strchr(key, ']');
			if (tmp != NULL) {
				*tmp = '\0';
				if (strcmp(&key[1], section) == 0) {
					if (verbose){
						printf(" ****** Reading section [%s]...\n", section);
					}
					found = 1;
				}
			}
		}

		if (verbose && !found){
			//printf(" * section [%s] not found!\n", section);
		}
	}//while()

#if 0
	char* dbg_ptr;
	int counter=0;
	Debug(DBG_CONF_FILE_READER, ("\nprinting [%s] section buffer:\n", section));
	for(dbg_ptr = buf; counter++ < dbg_len; dbg_ptr++){
		if(isprint(*dbg_ptr) || isspace(*dbg_ptr)){
			Debug(DBG_CONF_FILE_READER, ("%c", *dbg_ptr));
		}
	}
	Debug(DBG_CONF_FILE_READER, ("\n\n"));
#endif

	//Debug(DBG_CONF_FILE_READER, ("returning buf: %p\n", buf));

	return buf;
}

/*============================================================================
* Read a single record from the configuration file.
*	Read configuration file stripping comments and leading spaces until
*	non-empty line is read.  Copy it to the destination buffer.
*
* Return string length (incl. terminating zero) or 0 if end of file has been
* reached.
*/
int wancfglib_conf_file_reader::read_conf_record (FILE* file, char* key, int max_len)
{
	char buf[MAX_CFGLINE];		/* line buffer */

	//Debug(DBG_CONF_FILE_READER, ("%s(): key: %s\n", __FUNCTION__, key));

	while (fgets(buf, MAX_CFGLINE, file)) {

		char* str;
		int len;

		/* Strip leading spaces and tabs */
		for (str = buf; *str && strchr(" \t", *str); ++str){
			;
		}

		/* Skip comments and empty lines */
		len = strcspn(str, "#;\n\r");
		if (len) {
			str[len] = '\0';
			//Debug(DBG_CONF_FILE_READER, ("%s(): str: %s\n", __FUNCTION__, str));
			strlcpy(key, str, max_len);
			return len + 1;
		}
	}
	return 0;
}

/*============================================================================
* Build a list of channel definitions.
*/

int wancfglib_conf_file_reader::read_interfaces_section(
	FILE* file,
	wancfglib_chan_def & parent_chan_def_obj,
	char* parent_dev_name,
	int* current_line_number
	)
{
	int err = 0;
	char* conf;		/* -> section buffer */
	char* key;		/* -> current configuration record */
	int len = 0;	/* record length */
	char *token[MAX_TOKENS];
	char dev_name_backup[LEN_OF_DBG_BUFF]; 
	int local_line_number = 0;

	Debug(DBG_CONF_FILE_READER, ("%s()\n", __FUNCTION__));

	if(parent_dev_name != NULL){
		Debug(DBG_CONF_FILE_READER, ("%s(): 'parent_dev_name': %s\n", __FUNCTION__, parent_dev_name));
	}else{
		Debug(DBG_CONF_FILE_READER, ("%s(): 'parent_dev_name' is NULL\n", __FUNCTION__));
	}

	/* Read [interfaces] section */
	parent_chan_def_obj.tmp_conf = conf = read_conf_section(file, "interfaces");
	if (conf == NULL){
		ERR_DBG_OUT(("Failed to read '[interfaces]' section!\n"));
		return ERR_CONFIG;
	}

	/* For each record in [interfaces] section create channel definition structure.
	 * If channel is usedby STACK, read channels belonging to this stack. */
	for (key = conf; !err && *key; key += len) {
		int toknum;
		char *channel_sect;
		wancfglib_chan_def *list_el_chan_def = NULL;
		chan_def_t *chandef = NULL;
		wanif_conf_t *chanconf = NULL;

		len = strlen(key) + 1;

		Debug(DBG_CONF_FILE_READER, ("key: %s\n", key));

		//Start parsing at the line passed as a paramter to this function,
		//all the preceeding lines were already parsed.
		if(local_line_number != *current_line_number){
			Debug(DBG_CONF_FILE_READER, ("skipping line: %d, key: %s\n", local_line_number, key));
			local_line_number++;
			continue;
		}

		toknum = tokenize(key, token);
		if (toknum < 4){
			ERR_DBG_OUT(("Invalid number of tokens: %d, in key:%s. Expected at least 4!\n",
				toknum, key));
			return ERR_CONFIG;
		}
		/*
		for(indx = 0; indx < toknum; indx++){
			Debug(DBG_CONF_FILE_READER_AFT, ("token[%d]: %s\n", indx, token[indx]));
		}
		*/
		Debug(DBG_CONF_FILE_READER_AFT, ("\n ========================= \n"));

		//Use copy of the 'name' because 'strstr()' inserts nulls and it will break things.
		snprintf(dev_name_backup, sizeof(dev_name_backup), token[0]); 

		if(parent_dev_name != NULL){

			char parent_dev_name_backup[LEN_OF_DBG_BUFF]; 

			snprintf(parent_dev_name_backup, sizeof(parent_dev_name_backup), "%s", parent_dev_name);

			Debug(DBG_CONF_FILE_READER, ("searching '%s' in '%s'\n", parent_dev_name_backup, dev_name_backup));

			//Check this line belongs to the parent device.
			if(strstr(dev_name_backup, parent_dev_name_backup) == NULL){
				//If not, return from the recursive call. We hit interface belonging to another parent device.
				Debug(DBG_CONF_FILE_READER, ("this line does NOT belong to parent device: %s\n", parent_dev_name));
				free(conf);
				return 0;
			}else{
				Debug(DBG_CONF_FILE_READER, ("this line BELONGS to parent device: %s\n", parent_dev_name));
			}
		}  

		//allocate and initialize channel definition structure
		list_el_chan_def = new wancfglib_chan_def();
		chandef = &list_el_chan_def->if_data;
		chanconf = chandef->chanconf;

		snprintf(chandef->name, WAN_IFNAME_SZ, token[0]);

		//Very often token[2] is only a single byte, it is not enough
		//for storing a zero-terminated string (the addr), for that
		//reason always allocate a bigger bugger than the length of token[2].
		WANCFGLIB_MALLOC(char, chandef->addr, 20);
		snprintf(chandef->addr, 20, token[2]);

		//NOTE: 'atoi()' returns zero if it failes AND if number is actually zero!
		if(atoi(chandef->addr) == 0){

			Debug(DBG_CONF_FILE_READER, ("no valid 'chandef->addr', using 'list->get_size()': %d\n",
				parent_chan_def_obj.next_objects_list.size()));

			//1. There is nothing at all (blank space instead of a number).
			//   It is possible for a STACK or PPP interface where
			//   there is no "address" or "dlci number".
			//OR
			//2. For TTY zero is a *valid* number.
			//OR
			//3. For WANCONFIG_MFR there can be "auto" instead of the "dlci number".
			//   In this case there must be only one dlci/interface and '1' will mean "auto".
			//
			//   Current 'size' of the list will be actually the sequence number. Start from 1.
			//   									0+1	
			snprintf(chandef->addr, MAX_ADDR_STR_LEN, "%d", parent_chan_def_obj.next_objects_list.size() + 1);
		}

		if(toknum == 6){
			if(strcmp(token[4],"tty") == 0){
				strlcpy(chandef->addr, token[2], MAX_ADDR_STR_LEN);
			}
		}
		//print 'addr' after all the updates    
		Debug(DBG_CONF_FILE_READER, ("Final 'chandef->addr': %s\n", chandef->addr));

		if (toknum > 5){
			chanconf->config_id = wancfglib_get_config_id_from_profile(token[4]);
			if(chanconf->config_id == PROTOCOL_NOT_SET){
				ERR_DBG_OUT(("Invalid 'config_id': %s in profile!\n", token[4]));
				return ERR_CONFIG;
			}
		}

		chandef->usedby = strdup(token[3]);
		if(wancfglib_get_used_by_integer_value(token[3]) < 0){
			ERR_DBG_OUT(("Unsupported 'usedby' string: '%s' for interface: %s!\n", token[3], token[0]));
			return ERR_CONFIG;
		}

		channel_sect = strdup(token[0]);

		if (toknum > 5){
			//there is LIP layer
			char *section=strdup(token[5]);

			/* X25_SW */
			if (toknum > 7){
				if (toknum > 6)
					chandef->virtual_addr = strdup(token[6]);
				if (toknum > 7)
					chandef->real_addr = strdup(token[7]);
				if (toknum > 8)
					chandef->label = strdup(token[8]);
			}else{
				if (toknum > 6)
					chandef->label = strdup(token[6]);
			}

			Debug(DBG_CONF_FILE_READER_AFT,
				("calling 'chandef->conf_profile = read_conf_section(file, section: %s)'\n",
				section));

			//read the profile
			chandef->conf_profile = read_conf_section(file, section);
			free(section);

		}else{
			chandef->conf_profile = NULL;
		}//if(toknum > 5)

		chandef->conf = read_conf_section(file, channel_sect);	
		free(channel_sect);

		//Now, after configuration was read to a buffer, 
		//set all structures to real values.
		err = configure_chan(chandef, chanconf->config_id);
		if (err){
			return ERR_CONFIG;
		}

		Debug(DBG_CONF_FILE_READER, ("inserting interface: %s.\n", token[0]));

		parent_chan_def_obj.next_objects_list.push_back(list_el_chan_def);

		//initialize 'link' pointer to parent's 'link_defs'
		chandef->link = &parent_chan_def_obj.link_defs;

#if 0
		//FIXME: check for duplicate values when inserting new chan.
		//		for example two DLCI 16 is not allowed.
		{
			ERR_DBG_OUT(("Failed to insert interface %s to list!\n", token[0]));
			return ERR_CONFIG;
		}
#endif
		//advance the global 'current_line_number' for the next iteration
		Debug(DBG_CONF_FILE_READER, ("old current_line_number: %d\n", *current_line_number));
		(*current_line_number)++;
		Debug(DBG_CONF_FILE_READER, ("new current_line_number: %d\n", *current_line_number));

		//update the local one in order not skip the next line.
		local_line_number = *current_line_number;

		if(wancfglib_get_used_by_integer_value(chandef->usedby) == STACK){
			//the next interface (line in the file) we read belongs to this STACK,
			//not to the original 'parent_objects_list'.
			//NOTE: recursive call!!
			err = read_interfaces_section(
				file,
				*list_el_chan_def,
				token[0],
				current_line_number
				);
		}//if()

	}//for()

	return 0;
}

/*============================================================================
 * Set configuration parameter.
 *	Look up a keyword in configuration description table to find offset
 *	and data type of the configuration parameter into configuration data
 *	buffer.  Convert parameter to apporriate data type and store it the
 *	configuration data buffer.
 *
 *	Return:	0 - success
 *		1 - error
 *		-1 - not found
 */
int wancfglib_conf_file_reader::set_conf_param (char* key, char* val, key_word_t* dtab, void* conf, int max_len)
{
	unsigned int tmp = 0;

	/* Search a keyword in configuration definition table */
	for (; dtab->keyword && strcmp(dtab->keyword, key); ++dtab);
	
	if (dtab->keyword == NULL) return -1;	/* keyword not found */

	/* Interpert parameter value according to its data type */

	if (dtab->dtype == DTYPE_FILENAME) {
		//return read_data_file(val, (void*)((char*)conf + dtab->offset));
		printf("Warning: DTYPE_FILENAME not supported\n");
		return 0;//reading firmware files not supported.
	}
	
	/* FIXME: Add code that will parse Octasic Chip configuration file */
	if (dtab->dtype == DTYPE_OCT_FILENAME) {
		return 0;
	}
	if (dtab->dtype == DTYPE_OCT_CHAN_CONF) {
		//return read_oct_chan_config(key, val, (void*)((char*)conf + dtab->offset));
		printf("Warning: DTYPE_OCT_CHAN_CONF not supported\n");
		return 0;
	}

	if (verbose) printf(" * Setting %s to %s\n", key, val);

	if (dtab->dtype == DTYPE_STR) {
		strlcpy((char*)conf + dtab->offset, val, max_len-dtab->offset);
		return 0;
	}

	if (!(isdigit(*val) || (*val=='-' && isdigit(*(val+1)))) || 
	    strcmp(key, "TE_ACTIVE_CH") == 0 || strcmp(key, "ACTIVE_CH") == 0 || 
	    strcmp(key, "TE_LBO") == 0 || strcmp(key, "LBO") == 0 ||
	    strcmp(key, "FE_MEDIA") == 0 || strcmp(key, "MEDIA") == 0 || 
	    strcmp(key, "RBS_CH_MAP") == 0 || strcmp(key, "TE_RBS_CH") == 0 ||
	    strcmp(key, "TDMV_DCHAN") == 0){
		look_up_t* sym;
		unsigned int tmp_ch;

		strupcase(val);
		for (sym = sym_table;
		     sym->ptr && strcmp((char*)sym->ptr, val);
		     ++sym);
		     
		if (sym->ptr == NULL) {
			int ok_zero=0;

			/* TE1 */
			if (strcmp(key, "TE_ACTIVE_CH") && strcmp(key, "ACTIVE_CH") && 
			    strcmp(key, "RBS_CH_MAP") && strcmp(key, "TE_RBS_CH") &&
			    strcmp(key, "TDMV_DCHAN")) {	
				if (verbose) printf(" * invalid term %s ...\n", val);
				return 1;
			}

			if (strcmp(key, "TDMV_DCHAN") == 0) {
              	ok_zero=1;
			}

			/* TE1 Convert active channel string to UINT */
		 	tmp_ch = wancfglib_parse_active_channel(val);	
			if (tmp_ch == 0 && !ok_zero){
				if (verbose) printf("Illegal active channel range! %s ...\n", val);
				fprintf(stderr, "Illegal active channel range! min=1 max=31\n");
				return -1;
			}
			tmp = (unsigned int)tmp_ch;
		} else {
		 	tmp = sym->val;
		}
	} else {
		tmp = strtoul(val, NULL, 0);
	}
	/* SANITY CHECK */
	switch (dtab->dtype) {

	case DTYPE_INT:
		SIZEOFASSERT(dtab, sizeof(int));
		*(int*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_UINT:
		SIZEOFASSERT(dtab, sizeof(unsigned int));
		*(unsigned int*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_LONG:
		SIZEOFASSERT(dtab, sizeof(long));
		*(long*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_ULONG:
		SIZEOFASSERT(dtab, sizeof(unsigned long));
		*(unsigned long*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_SHORT:
		SIZEOFASSERT(dtab, sizeof(short));
		*(short*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_USHORT:
		SIZEOFASSERT(dtab, sizeof(unsigned short));
		*(ushort*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_CHAR:
		SIZEOFASSERT(dtab, sizeof(char));
		*(char*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_UCHAR:
		SIZEOFASSERT(dtab, sizeof(unsigned char));
		*(unsigned char*)((char*)conf + dtab->offset) = tmp;
		break;

	case DTYPE_PTR:
		*(void**)((char*)conf + dtab->offset) = (void*)tmp;
		break;
	}
	return 0;
}

/*============================================================================
* Configure WAN logical channel.
*/

int wancfglib_conf_file_reader::configure_chan (chan_def_t* chandef, int id)
{
	int err = 0;
	int len = 0;
	char* conf_rec;
	char *token[MAX_TOKENS];

	Debug(DBG_CONF_FILE_READER, ("%s(id: %d)\n", __FUNCTION__, id));

	/* Prepare configuration data */
	strncpy(chandef->chanconf->name, chandef->name, WAN_IFNAME_SZ);

	if (chandef->addr){
		strncpy(chandef->chanconf->addr, chandef->addr, WAN_ADDRESS_SZ);
	}

	if (chandef->usedby){
		strncpy(chandef->chanconf->usedby, chandef->usedby, USED_BY_FIELD);
	}

	if(chandef->label){
		memcpy(chandef->chanconf->label, chandef->label, WAN_IF_LABEL_SZ);
	}

	Debug(DBG_CONF_FILE_READER, ("chandef->conf ptr: 0x%p\n", chandef->conf));

	//If there is any information in the [interface] section, get it.
	//Some interfaces have no information.
	if (chandef->conf){

		if(id == 0){
			//This the hardware interface, so we can set id to WANCONFIG_AFT_TE1.
			id = WANCONFIG_AFT_TE1;
		}

		key_word_t* conf_table = (key_word_t*)lookup(id, conf_if_def_tables);

		if(!conf_table){
			printf("Warning: Failed to find configuration table for id: %d. Setting to 'xilinx_if_conftab'.\n", id);
			conf_table = xilinx_if_conftab;
		}

		for (conf_rec = chandef->conf; *conf_rec; conf_rec += len) {
			int toknum;

			len = strlen(conf_rec) + 1;
			toknum = tokenize(conf_rec, token);
			if (toknum < 2) {
				return -EINVAL;
			}

			// Look up a keyword first in common, then media-specific
			// configuration definition tables.
			strupcase(token[0]);
			if (set_conf_param(token[0], token[1], chan_conftab, chandef->chanconf, sizeof(wanif_conf_t))){

				if(set_conf_param(token[0], token[1], conf_table, &chandef->chanconf->u, sizeof(chandef->chanconf->u))){
					//not found in 'conf_if_def_tables' too !!
					ERR_DBG_OUT(("Error: Invalid parameter: %s\n", token[0]));
					return ERR_CONFIG;
				}
			}else{
				if (strcmp(token[0], "ACTIVE_CH") == 0){
					strlcpy(chandef->active_ch, token[1], 50);
				}
			}
		}//for()

		Debug(DBG_CONF_FILE_READER, ("Current 'chandef->chanconf->config_id': %d\n", chandef->chanconf->config_id));

		switch(wancfglib_get_used_by_integer_value(chandef->usedby))
		{
		case TDM_VOICE:
			Debug(DBG_CONF_FILE_READER, ("TDM_VOICE channel!!\n"));
			chandef->chanconf->config_id = PROTOCOL_TDM_VOICE;
			break;
		case TDM_VOICE_API:
			Debug(DBG_CONF_FILE_READER, ("TDM_VOICE_API channel!!\n"));
			chandef->chanconf->config_id = PROTOCOL_TDM_VOICE_API;
			break;
		case TDM_SPAN_VOICE_API:
			Debug(DBG_CONF_FILE_READER, ("TDM_SPAN_VOICE_API channel!!\n"));
			chandef->chanconf->config_id = PROTOCOL_TDM_SPAN_VOICE_API;
			break;
		case TDM_CHAN_VOICE_API:
			Debug(DBG_CONF_FILE_READER, ("TDM_CHAN_VOICE_API channel!!\n"));
			chandef->chanconf->config_id = PROTOCOL_TDM_CHAN_VOICE_API;
			break;
		case API_LEGACY:
			Debug(DBG_CONF_FILE_READER, ("API_LEGACY channel!!\n"));
			chandef->chanconf->config_id = PROTOCOL_API_LEGACY;
			break;
		}

		Debug(DBG_CONF_FILE_READER, ("Final 'chandef->chanconf->config_id': %d.\n", chandef->chanconf->config_id));

		len=0;

	}else{
		Debug(DBG_CONF_FILE_READER, ("chandef->conf == NULL!!\n"));
	}	

	Debug(DBG_CONF_FILE_READER, ("chandef->conf_profile ptr: 0x%p\n", chandef->conf_profile));

	//if there is anything in [profile] section, get it
	if (chandef->conf_profile){
			
		wanif_conf_t *chanconf = chandef->chanconf;

		Debug(DBG_CONF_FILE_READER_AFT, ("chandef->addr: %s\n", chandef->addr));	
		Debug(DBG_CONF_FILE_READER_AFT, ("chanconf ptr: 0x%p, chanconf->config_id: %d (%s)\n",
			chanconf, chanconf->config_id, wancfglib_get_protocol_string(chanconf->config_id)));	
#if USE_ANNEXG_TABLE
		key_word_t *conf_annexg_table = (key_word_t*)lookup(wancfglib_config_id_to_annexg_id(chanconf->config_id), conf_annexg_def_tables);
#else
		key_word_t *conf_annexg_table = (key_word_t*)lookup(chanconf->config_id, conf_def_tables);
#endif

		if(!conf_annexg_table){
			ERR_DBG_OUT(("Protocol %d(%s) MUST NOT have a profile section!\n", 
				chanconf->config_id, wancfglib_get_protocol_string(chanconf->config_id)));
			return ERR_CONFIG;
		}

		for(conf_rec = chandef->conf_profile; *conf_rec; conf_rec += len) {

			int toknum;

			len = strlen(conf_rec) + 1;
			toknum = tokenize(conf_rec, token);

			Debug(DBG_CONF_FILE_READER_AFT, ("toknum: %d\n", toknum));	

			if (toknum != 2) {
				return -EINVAL;
			}

			/* Look up a keyword first in common, then media-specific
			 * configuration definition tables.
			 */
			strupcase(token[0]);

			Debug(DBG_CONF_FILE_READER_AFT, ("token[0]: %s\n", token[0]));
			Debug(DBG_CONF_FILE_READER_AFT, ("token[1]: %s\n", token[1]));

			if(set_conf_param(token[0], token[1], conf_annexg_table, &chandef->chanconf->u, sizeof(chandef->chanconf->u))) {

				if (set_conf_param(token[0], token[1], chan_conftab, chandef->chanconf, sizeof(wanif_conf_t))) {

					ERR_DBG_OUT(("Invalid parameter %s\n", token[0]));
					return ERR_CONFIG;
				}

			}

		}//for()

	}else{
		Debug(DBG_CONF_FILE_READER, ("chandef->conf_profile == NULL!!\n"));
	}

	return err;
}

////////////////////////////////////////////////////////////////////////////////////////////////
//Global (non-class-member) functions.
////////////////////////////////////////////////////////////////////////////////////////////////
void init_first_time_tokens(char **token){
	int i;

	for (i=0;i<MAX_TOKENS;i++){
		token[i]=NULL;
	}
}

/*============================================================================
* Tokenize string.
*	Parse a string of the following syntax:
*		<tag>=<arg1>,<arg2>,...
*	and fill array of tokens with pointers to string elements.
*
*	Return number of tokens.
*/
void init_tokens(char **token){
	int i;

	for (i=0;i<MAX_TOKENS;i++){
		token[i]=NULL;
	}
}

/* Bug Fix by Ren�Scharfe <l.s.r@web.de>
* removed strtok
*/
int tokenize (char *str, char **tokens)
{
	int cnt = 0;
	char *tok;

	init_tokens(tokens);

	if (!str){
		return 0;
	}

	tok = strchr(str, '=');
	if (!tok)
		return 0;

	*tok='\0';

	tokens[cnt] = str;
	str=++tok;

	while (tokens[cnt] && (cnt < MAX_TOKENS-1)) {

		tokens[cnt] = str_strip(tokens[cnt], " \t");

		if ((tok = strchr(str, ',')) == NULL){
			tokens[++cnt] = str_strip(str, " \t");
			goto end_tokenize;
		}

		*tok='\0';

		tokens[++cnt] = str_strip(str, " \t");
		str=++tok;
	}

end_tokenize:
	return ++cnt;
}

/*============================================================================
* Strip leading and trailing spaces off the string str.
*/
char* str_strip (char* str, char* s)
{
	char* eos = str + strlen(str);		/* -> end of string */

	while (*str && strchr(s, *str))
		++str;				/* strip leading spaces */
	while ((eos > str) && strchr(s, *(eos - 1)))
		--eos;				/* strip trailing spaces */
	*eos = '\0';
	return str;
}

/*============================================================================
* Uppercase string.
*/
char* strupcase (char* str)
{
	char* s;

	for(s = str; *s; ++s) *s = toupper(*s);
	return str;
}

/*============================================================================
* Get a pointer from a look-up table.
*/
void* lookup (int val, look_up_t* table)
{
	while(table->val && (table->val != (uint)val)){
		table++;
	}
	return table->ptr;
}

/*============================================================================
* Look up a symbolic name in name table.
*	Return a numeric value associated with that name or zero, if name was
*	not found.
*/
int name2val (char* name, look_up_t* table)
{
	while (table->ptr && strcmp(name, (char*)table->ptr)) table++;
	return table->val;
}

/*============================================================================
* Convert decimal string to unsigned integer.
* If len != 0 then only 'len' characters of the string are converted.
*/
unsigned int dec_to_uint (unsigned char* str, int len)
{
	unsigned val;

	if (!len) len = strlen((char*)str);
	for (val = 0; len && is_digit(*str); ++str, --len)
		val = (val * 10) + (*str - (unsigned)'0')
		;
	return val;
}

/****** End *****************************************************************/
